#!/usr/bin/env python

__author__ = "Philipp Blandfort"


# Notes:
# In this model, a chi-squared distribution is used to model h.
# Variable names are different from names in thesis!
# R_0 is not used anymore. (It is assumed that this is equal to R_ges.)
# psc_factor has been removed from the parameters.
# ca_baseline is set to 0.
# Optimization method is now always L-BFGS-B.


#===========================================================================================================================
# IMPORTS
#===========================================================================================================================

from scipy import *
from numpy import *
from scipy.stats import chi2
from matplotlib.pyplot import *
from scipy.optimize import fmin_l_bfgs_b
from scipy.integrate import odeint
from scipy.integrate import ode


#===========================================================================================================================
# DISTANCE FUNCTIONS
# Functions to use to compute vector distances (to determine objective function value).
#===========================================================================================================================

def oneDist(v1, v2):
    return sum(abs(v1-v2))

def twoDist(v1, v2):
    return sum((v1-v2)**2)


#===========================================================================================================================
# THE MAIN CLASS
#===========================================================================================================================

class ODEModel:
    #=======================================================================================================================
    # CLASS CONSTANTS AND SETTINGS
    #=======================================================================================================================

    # parameters (note that the order is relevant!)
    keys = ['b', 'w_bar', 'alpha', 'tau', 'delta', 'k', 'gamma', 'h_amp', 'h_width', 'h_delay', 'R_ges', 's_min', 's_max', 
        'mu_factor', 'ca_jump', 'ca_decay', 'ca_delay']

    # boundary conditions (order is relevant!)
    boundary_keys = ['N_in_0', 'N_v_0', 'N_out_0', 'v_0']

    # bounds for the parameters (for parameter optimization)
    bounds = {'b': (0, None), 'w_bar': (0, None), 'alpha': (0, None),
        'tau': (0, None), 'delta': (0, None), 'k': (0, None),
        'gamma': (0, None), 'h_amp': (0, None), 'h_width': (0, None), 'h_delay': (0, None), 'R_ges': (0, None),
        's_min': (0, None), 's_max': (1., None), 'mu_factor': (0, None), 'ca_jump': (0, None), 'ca_decay': (0, None),
        'ca_delay': (0, None), 'N_in_0': (0, None), 'N_v_0': (0, None), 'N_out_0': (0, None),
        'v_0': (0, None)}

    # path to the data
    data_path = (lambda self, frequency: 'data/'+str(frequency)+'Hz.npz')

    # default parameters
    # see initialization

    # number of state variables
    variable_number = 5 # N_in, N_v, N_out, v, R


    #=======================================================================================================================
    # INITIALIZATION
    #=======================================================================================================================

    def __init__(self, boundary_conditions = [1., 0., 0., 5.], frequency = 1, data_part = slice(9900,10300),
        distance_fct = twoDist):
        # boundary conditions (these values can be changed by setBoundaryConditions())
        self.N_in_0 = boundary_conditions[0]
        self.N_v_0 = boundary_conditions[1]
        self.N_out_0 = boundary_conditions[2]
        self.v_0 = boundary_conditions[3]

        # load the data
        self.loadData(frequency, data_part)

        # initial parameter values used for optimiziation
        self.p = {'b': 7., 'w_bar': 10., 'alpha': 1., 'tau': .01, 'delta': .1, 'k': .03, 'gamma': .07, 'h_amp': .5,
            'h_width': 10., 'R_ges': 15., 's_min': 2., 's_max': 5.,
            'h_delay': 5., 'mu_factor': .05, 'ca_jump': 7., 'ca_decay': .07, 'ca_delay': 7}

        # distance function used to compute the error
        self.distance = distance_fct

        # initial keyword arguments for odeint()
        self.odekwargs = {}


    #=======================================================================================================================
    # GETTERS AND SETTERS
    #=======================================================================================================================

    def setBoundaryConditions(self, N_in_0 = 1., N_v_0 = 0., N_out_0 = 0., v_0 = .5):
        '''Set boundary conditions for the ODE model.'''
        self.N_in_0 = N_in_0
        self.N_v_0 = N_v_0
        self.N_out_0 = N_out_0
        self.v_0 = v_0

    def getBoundaryConditions(self):
        '''Get the boundary conditions of the model.

        @return boundary conditions as dictionary'''
        return {'N_in_0': self.N_in_0, 'N_v_0': self.N_v_0, 'N_out_0': self.N_out_0, 'v_0': self.v_0}

    def getData(self):
        '''Get the input and desired output of the system.'''
        return self.data, self.target

    def setParams(self, params):
        '''Set the initial parameter values that are used for optimization.'''
        assert len(params)>=len(self.keys), 'Not enough parameters!'
        self.p = params

    def getParams(self):
        '''Get the initial parameter values that are used for optimization.'''
        return self.p

    def setDistanceFct(self, distance_fct):
        '''Set the distance function used to compute the error.

        @param distance_fct function taking two vectors as parameters and returning their distance'''
        self.distance = distance_fct


    #=======================================================================================================================
    # LOADING DATA
    #=======================================================================================================================

    def loadData(self, frequency = 1, part = slice(9900,10300)):
        npzfile = np.load(self.data_path(frequency))
        stimulus = npzfile['stimulus'][part]
        target = npzfile['output'][part]
        self.data = stimulus
        self.target = target


    #=======================================================================================================================
    # UTILITY FUNCTIONS
    # Utility functions for parameter handling. (Conversion between list and dictionaries.)
    #=======================================================================================================================

    def valueDicToList(self, dic_values):
        '''Convert a dictionary of values (parameters and boundary conditions) to a simple list.'''
        if len(dic_values)>=len(self.keys)+len(self.boundary_keys):
           return self.paramDicToList(dic_values) + self.boundaryDicToList(dic_values)
        else:
            return self.paramDicToList(dic_values)

    def valueListToDic(self, value_list):
        '''Convert a list of values to a dictionary.

        @param value_list list of values for the parameters and boundary conditions'''
        assert len(value_list)>=len(self.keys)+len(self.boundary_keys), 'Not enough parameters!'
        dic = {}
        for i, value in enumerate(value_list):
            if i < len(self.keys):
                dic[self.keys[i]] = value
            else:
                dic[self.boundary_keys[i-len(self.keys)]] = value
        return dic

    def paramDicToList(self, dic_params):
        '''Convert a parameter dictionary to a simple list.'''
        l = []
        for param_name in self.keys:
            l.append(dic_params[param_name])
        return l

    def paramListToDic(self, param_list):
        '''Convert a list of parameters to a dictionary.

        @param param_list list of values for the parameters'''
        assert len(param_list)>=len(self.keys), 'Not enough parameters!'
        dic = {}
        for i, value in enumerate(param_list):
            if i < len(self.keys):
                dic[self.keys[i]] = value
            else:
                break
        return dic

    def boundaryDicToList(self, dic_values):
        '''Convert a dictionary with boundary condition values to a simple list.'''
        l = []
        for param_name in self.boundary_keys:
            l.append(dic_values[param_name])
        return l

    def paramIndex(self, param_name):
        '''Get the position of a certain parameter in the list of keys.

        @param param_name name of the parameter'''
        return self.keys.index(param_name)


    #=======================================================================================================================
    # MODEL FUNCTIONS
    # Functions that are used to solve the ODE model.
    #=======================================================================================================================

    def lin_psc(self, rocc, factor=100.):
        '''Calculate the postsynaptic current, assuming
        linear dependency between the number of occupied receptors and the PSC.

        @param rocc occupied receptors
        @param factor coefficient of linear dependency
        @return the resulting postsynaptic current (in pA)'''
        return factor*rocc

    def calciumConcentration(self, t, a=.07, b=7, Ca0=0., delay=5):
        '''Estimate the calcium concentration in the presynaptic terminal at time t.

        @param t time
        @param a rate of decay
        @param b change in amplitude as response to stimulus
        @param Ca0 the baseline concentration
        @param delay response time'''
        summand = 0
        for ti in where(self.data)[0]:
            if t >= ti + delay:
                summand += exp(-a * (t - delay - ti))
        return Ca0 + b*summand

    def mu(self, factor=.05, *calcium_args):
        '''Determine the function mu(t).

        (Assuming linear dependency between calcium concentration and vesicle release rate.)

        @param data input data as an array where data[i] contains the input at time t_i
        @param factor coefficient of linear dependency
        @param calcium_args additional arguments to pass to the calciumConcentration() function
        @return function that takes the time index t as first and the vesicle size s as second argument'''
        return (lambda t: maximum(0., factor*self.calciumConcentration(t, *calcium_args)))

    def h(self, t, amplitude, width, delay, f=3):
        summand = 0
        for ti in where(self.data)[0]:
            if t >= ti + delay:
                summand += chi2.pdf((t - ti - delay)*width*1./100, f)
        return amplitude*summand

    def output(self, *params):
        '''Calculate the output of our system.'''
        assert len(params)>=len(self.keys)
        return self.lin_psc(params[self.paramIndex('R_ges')] - self.solve_system(*params)[:,4])

    def f(self, y, t, params):
        '''Compute the time derivative at time t. Given the parameters params and current state y.'''
        assert len(params)>=len(self.keys)

        # Model parameters
        b, w_bar, alpha, tau, delta, k, gamma, h_amp, h_width, h_delay, R_ges, s_min, s_max = params[:13]

        # Function parameters
        mu_factor, ca_jump, ca_decay, ca_delay = params[13:17]

        h_fct = (lambda t: self.h(t, h_amp, h_width, h_delay))
        ca_baseline = 0.
        mu_fct = self.mu(mu_factor, ca_decay, ca_jump, ca_baseline, ca_delay)

        return [b * (w_bar - y[0]) * ((w_bar - y[0])>0) + alpha * y[2] - tau * y[0] * 1./(s_max - s_min) * (s_max * y[3] - y[1]), \
                y[0] * tau * 1./(s_max - s_min) * (s_max * y[3] - y[1]) - mu_fct(t) * y[1], \
                mu_fct(t) * y[1] - k * y[2] * y[4] - (alpha + delta) * y[2] + gamma * (R_ges - y[4]), \
                y[0] * tau * h_fct(t) - mu_fct(t) * y[3], \
                - k * y[4] * y[2] + gamma * (R_ges - y[4])]

    def f_jacobian(self, y, t, params):
        '''Compute the Jacobian (w.r.t. y) of the time derivative function at time t.
        Given the parameters params and current system state y.'''
        assert len(params)>=len(self.keys)

        # Model parameters
        b, w_bar, alpha, tau, delta, k, gamma, h_amp, h_width, h_delay, R_ges, s_min, s_max = params[:13]

        # Function parameters
        mu_factor, ca_jump, ca_decay, ca_delay = params[13:17]

        h_fct = (lambda t: self.h(t, h_amp, h_width, h_delay))
        ca_baseline = 0.
        mu_fct = self.mu(mu_factor, ca_decay, ca_jump, ca_baseline, ca_delay)

        return [[-b * (y[0]<w_bar) - tau * 1./(s_max - s_min) * (s_max * y[3] - y[1]),
                tau * y[0] * 1./(s_max - s_min), alpha, -tau*y[0]*1./(s_max - s_min)*s_max, 0],
            [tau * 1./(s_max - s_min) * (s_max * y[3] - y[1]), -y[0]*tau*1./(s_max - s_min)-mu_fct(t), 0, y[0]*tau*1./(s_max - s_min)*s_max, 0],
            [0, mu_fct(t), -k*y[4] - (alpha+delta), 0, -k*y[2] + gamma],
            [tau * h_fct(t), 0, 0, -mu_fct(t), 0],
            [0, 0, -k*y[4], 0, -k*y[2] - gamma]]

    def solve_system(self, *params):
        '''Calculate the behaviour of all variables for a given input.
        (Note that this function uses global variables for boundary conditions.)

        @param data input data as an array where data[i] contains the input at time t_i
        @param params parameter values for the model
        @return array with the values of the system variables'''        
        # Initialization
        output = zeros((len(self.data),self.variable_number))

        # Applying the boundary conditions
	R_0 = self.paramListToDic(params)['R_ges']
        if len(params) >= len(self.keys) + len(self.boundary_keys):
            x = array(list(params[-len(self.boundary_keys):]) + [R_0])
            return odeint((lambda y,t: self.f(y,t,params)), x, range(len(self.data)), **self.odekwargs)
        else:
            x = array([self.N_in_0, self.N_v_0, self.N_out_0, self.v_0, R_0])
            return odeint((lambda y,t: self.f(y,t,params)), x, range(len(self.data)), Dfun=(lambda y,t: self.f_jacobian(y,t,params)), **self.odekwargs)


    #=======================================================================================================================
    # EVALUATION
    # Computing how well the model output fits the given data.
    #=======================================================================================================================

    def error(self, params = None):
        '''Calculate the objective function value given a certain configuration of model parameters.

        @param params parameter values for the model
        @param oneDist a function taking two vectors as arguments and returning the distance between them
        @return error for the given parameters'''
        if params is None:
            params = self.valueDicToList(self.p)
        return self.distance(self.output(*params), self.target)


    #=======================================================================================================================
    # VISUALIZATION
    # Displaying output and behaviour of individual variables.
    #=======================================================================================================================

    def plotPerformance(self, params = None):
        '''Plot the input, output and desired output for the given choice of parameters. Note that the input is scaled.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        plot(array(range(len(self.data)))*1./10, self.data * amax(self.target)/amax(self.data), label="input", alpha=.6)
        plot(array(range(len(self.data)))*1./10, self.output(*params), label="output")
        plot(array(range(len(self.data)))*1./10, self.target, label="target")
        legend()

    def plotVariableBehaviour(self, params = None):
        '''Plot the behaviour of the different system variables, given some input and a list of parameters.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        variables = self.solve_system(*params)
        paramDic = self.paramListToDic(params)
        ca_baseline = 0.
        plot(array(range(len(self.data)))*1./10, [self.calciumConcentration(t, paramDic['ca_decay'], paramDic['ca_jump'], ca_baseline,\
            paramDic['ca_delay']) for t in range(len(self.data))], label="$Ca^{2+}$")
        plot(array(range(len(self.data)))*1./10, variables[:,0], label="$N_{in}$")
        plot(array(range(len(self.data)))*1./10, variables[:,1], label="$N_v$")
        plot(array(range(len(self.data)))*1./10, variables[:,2], label="$N_{out}$")
        plot(array(range(len(self.data)))*1./10, variables[:,3], label="$v$")
        plot(array(range(len(self.data)))*1./10, variables[:,4], label="$R$")
        legend()

    def plotBehaviour(self, params = None):
        '''Plot performance of the model and variable behaviour.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        figsize(15, 5)
        subplot(1, 2, 1)
        self.plotPerformance(params)

        subplot(1, 2, 2)
        self.plotVariableBehaviour(params)

        # set figsize back to normal for later plots
        figsize(9,5)


    #=======================================================================================================================
    # OPTIMIZATION
    # Finding optimal parameters for the model.
    #=======================================================================================================================

    def optimize(self, display = True, **kwargs):
        '''Find the optimal parameters.

        @param display If set to True, the optimal parameters are printed out and the performance is plotted.
        @return the (hopefully) optimal parameters'''
        if len(self.p) >= len(self.keys) + len(self.boundary_keys):
            b = self.valueDicToList(self.bounds)
        else:
            b = self.paramDicToList(self.bounds)
        result = fmin_l_bfgs_b((lambda params: self.error(params)), self.valueDicToList(self.p), bounds=b,
            approx_grad=True, **kwargs)
        opt_p = result[0]

        if display:
            if len(self.p) >= len(self.keys) + len(self.boundary_keys):
                print 'optimal parameters:', self.valueListToDic(opt_p)
            else:
                print 'optimal parameters:', self.paramListToDic(opt_p)

            figsize(15, 5)
            subplot(1, 2, 1)
            self.plotPerformance(opt_p)

            subplot(1, 2, 2)
            self.plotVariableBehaviour(opt_p)

            # set figsize back to normal for later plots
            figsize(9,5)

        return result
