#!/usr/bin/env python

__author__ = "Philipp Blandfort"


# NOTES:
# Variable names are different from names in thesis!
# R_0 is not used anymore. (It is assumed that this is equal to R_ges.)
# psc_factor has been removed from the parameters.
# ca_baseline is set to 0.
# m is not modeled (because it becomes relevant only after several seconds of stimulus) and assumed to be 0 all the time.
# -> r not used anymore.
# Using saturation effects.

#===========================================================================================================================
# IMPORTS
#===========================================================================================================================

from scipy import *
from numpy import *
from matplotlib.pyplot import *
from scipy.optimize import fmin_l_bfgs_b
from scipy.integrate import odeint
from scipy.integrate import ode


#===========================================================================================================================
# DISTANCE FUNCTIONS
# Functions to use to compute vector distances (to determine objective function value).
#===========================================================================================================================

def oneDist(v1, v2):
    return sum(abs(v1-v2))

def twoDist(v1, v2):
    return sum((v1-v2)**2)


#===========================================================================================================================
# THE MAIN CLASS
#===========================================================================================================================

class ODEModel:
    #=======================================================================================================================
    # CLASS CONSTANTS AND SETTINGS
    #=======================================================================================================================

    # parameters (note that the order is relevant!)
    keys = ['b', 'w_bar', 'alpha', 'delta', 'k', 'gamma', 'R_ges', 'chi', 'phi', 'z', 'p',
        'mu_factor', 'ca_jump', 'ca_decay', 'ca_delay']

    # boundary conditions (order is relevant!)
    boundary_keys = ['n_p_0', 'n_r_0', 'N_p_0', 'N_r_0', 'N_in_0', 'N_out_0']

    # bounds for the parameters (for parameter optimization)
    bounds = {'b': (0, None), 'w_bar': (0, None), 'alpha': (0, None),
        'delta': (0, None), 'k': (0, None), 'gamma': (0, None), 'R_ges': (0, None),
        'mu_factor': (0, None), 'ca_jump': (0, None), 'ca_decay': (0, None),
        'ca_delay': (0, None), 'chi': (0, None), 'phi': (0, None), 'z': (0, None), 'p': (0, None),
        'n_p_0': (0, None), 'n_r_0': (0, None), 'N_p_0': (0, None), 'N_r_0': (0, None),
        'N_in_0': (0, None), 'N_out_0': (0, None), 'R_0': (0, None), 'm_0': (0, None)}

    # path to the data
    data_path = (lambda self, frequency: 'data/'+str(frequency)+'Hz.npz')

    # default parameters
    # see initialization

    # number of state variables
    variable_number = 7 # n_p, n_r, N_p, N_r, N_in, N_out, R


    #=======================================================================================================================
    # INITIALIZATION
    #=======================================================================================================================

    def __init__(self, boundary_conditions = [10., 10., 0., 0., 5., 0.], frequency = 1,
        data_part = slice(9900,10300), distance_fct = twoDist):
        # boundary conditions
        self.setBoundaryConditions(*boundary_conditions)

        # load the data
        self.loadData(frequency, data_part)

        # initial parameter values used for optimiziation
        self.p = {'b': .5, 'w_bar': 20., 'alpha': 1., 'delta': .1, 'k': .03, 'gamma': .07, 'R_ges': 15.,
            'mu_factor': .05, 'ca_jump': 7., 'ca_decay': .07, 'ca_delay': 7,
            'chi': .05, 'phi': .1, 'z': 5., 'p': 15.}

        # distance function used to compute the error
        self.distance = distance_fct

        # initial keyword arguments for odeint()
        self.odekwargs = {}


    #=======================================================================================================================
    # GETTERS AND SETTERS
    #=======================================================================================================================

    def setBoundaryConditions(self, n_p_0 = 10., n_r_0 = 10., N_p_0 = 0., N_r_0 = .0, N_in_0 = 5., N_out_0 = 0.):
        '''Set boundary conditions for the ODE model.

        Note that these values are only relevant if self.p doesn't contain values for the boundary conditions.'''
        self.n_p_0 = n_p_0
        self.n_r_0 = n_r_0
        self.N_p_0 = N_p_0
        self.N_r_0 = N_r_0
        self.N_in_0 = N_in_0
        self.N_out_0 = N_out_0

    def getBoundaryConditions(self):
        '''Get the boundary conditions of the model.

        @return boundary conditions as dictionary'''
        return {'n_p_0': self.n_p_0, 'n_r_0': self.n_r_0, 'N_p_0': self.N_p_0, 'N_r_0': self.N_r_0,
            'N_in_0': self.N_in_0, 'N_out_0': self.N_out_0}

    def getData(self):
        '''Get the input and desired output of the system.'''
        return self.data, self.target

    def setParams(self, params):
        '''Set the initial parameter values that are used for optimization.'''
        assert len(params)>=len(self.keys), 'Not enough parameters!'
        self.p = params

    def getParams(self):
        '''Get the initial parameter values that are used for optimization.'''
        return self.p

    def setDistanceFct(self, distance_fct):
        '''Set the distance function used to compute the error.

        @param distance_fct function taking two vectors as parameters and returning their distance'''
        self.distance = distance_fct


    #=======================================================================================================================
    # LOADING DATA
    #=======================================================================================================================

    def loadData(self, frequency = 1, part = slice(9900,10300)):
        npzfile = np.load(self.data_path(frequency))
        stimulus = npzfile['stimulus'][part]
        target = npzfile['output'][part]
        self.data = stimulus
        self.target = target


    #=======================================================================================================================
    # UTILITY FUNCTIONS
    # Utility functions for parameter handling. (Conversion between list and dictionaries.)
    #=======================================================================================================================

    def valueDicToList(self, dic_values):
        '''Convert a dictionary of values (parameters and boundary conditions) to a simple list.'''
        if len(dic_values)>=len(self.keys)+len(self.boundary_keys):
           return self.paramDicToList(dic_values) + self.boundaryDicToList(dic_values)
        else:
            return self.paramDicToList(dic_values)

    def valueListToDic(self, value_list):
        '''Convert a list of values to a dictionary.

        @param value_list list of values for the parameters and boundary conditions'''
        #assert len(value_list)>=len(self.keys)+len(self.boundary_keys), 'Not enough parameters!'
        #TODO use other assertion?
        dic = {}
        for i, value in enumerate(value_list):
            if i < len(self.keys):
                dic[self.keys[i]] = value
            else:
                dic[self.boundary_keys[i-len(self.keys)]] = value
        return dic

    def paramDicToList(self, dic_params):
        '''Convert a parameter dictionary to a simple list.'''
        l = []
        for param_name in self.keys:
            l.append(dic_params[param_name])
        return l

    def paramListToDic(self, param_list):
        '''Convert a list of parameters to a dictionary.

        @param param_list list of values for the parameters'''
        assert len(param_list)>=len(self.keys), 'Not enough parameters!'
        dic = {}
        for i, value in enumerate(param_list):
            if i < len(self.keys):
                dic[self.keys[i]] = value
            else:
                break
        return dic

    def boundaryDicToList(self, dic_values):
        '''Convert a dictionary with boundary condition values to a simple list.'''
        l = []
        for param_name in self.boundary_keys:
            l.append(dic_values[param_name])
        return l

    def paramIndex(self, param_name):
        '''Get the position of a certain parameter in the list of keys.

        @param param_name name of the parameter'''
        return self.keys.index(param_name)


    #=======================================================================================================================
    # MODEL FUNCTIONS
    # Functions that are used to solve the ODE model.
    #=======================================================================================================================

    def lin_psc(self, rocc, factor=100.):
        '''Calculate the postsynaptic current, assuming
        linear dependency between the number of occupied receptors and the PSC.

        @param rocc occupied receptors
        @param factor coefficient of linear dependency
        @return the resulting postsynaptic current (in pA)'''
        return factor*rocc

    def calciumConcentration(self, t, a=.07, b=7, Ca0=0., delay=5):
        '''Estimate the calcium concentration in the presynaptic terminal at time t.

        @param t time
        @param a rate of decay
        @param b change in amplitude as response to stimulus
        @param Ca0 the baseline concentration
        @param delay response time'''
        summand = 0
        for ti in where(self.data)[0]:
            if t >= ti + delay:
                summand += exp(-a * (t - delay - ti))
        return Ca0 + b*summand

    def mu(self, factor=.05, *calcium_args):
        '''Determine the function mu(t).

        @param data input data as an array where data[i] contains the input at time t_i
        @param factor coefficient of linear dependency
        @param calcium_args additional arguments to pass to the calciumConcentration() function
        @return function that takes the time index t as first and the vesicle size s as second argument'''
        # assuming linear dependency between calcium concentration and vesicle release rate
        return (lambda t: maximum(0., factor*self.calciumConcentration(t, *calcium_args)))

    def output(self, *params):
        '''Calculate the output of our system.'''
        assert len(params)>=len(self.keys)
        return self.lin_psc(params[self.paramIndex('R_ges')] - self.solve_system(*params)[:,6])

    def f(self, y, t, params):
        '''Compute the time derivative at time t. Given the parameters params and current state y.'''
        assert len(params)>=len(self.keys)

        # Model parameters
        b, w_bar, alpha, delta, k, gamma, R_ges, chi, phi, z, p = params[:11]

        # Function parameters
        mu_factor, ca_jump, ca_decay, ca_delay = params[11:15]

        ca_baseline = 0.
        mu_fct = self.mu(mu_factor, ca_decay, ca_jump, ca_baseline, ca_delay)

        y_7_delta_m = 0 # usually we would need a delay here

        #chi = chi * (p - y[0]) * ((p - y[0])>0)
        #phi_p = phi * (z * y[0] - y[2]) * ((z * y[0] - y[2])>0)
        #phi_r = phi * (z * y[1] - y[3]) * ((z * y[1] - y[3])>0)
        return [chi * (p - y[0]) * ((p - y[0])>0) * y[1] - mu_fct(t) * y[0],
            - chi * (p - y[0]) * ((p - y[0])>0) * y[1],
            chi * (p - y[0]) * ((p - y[0])>0) * y[3] - mu_fct(t) * y[2] + phi * (z * y[0] - y[2]) * ((z * y[0] - y[2])>0) * y[0] * y[4],
            phi * (z * y[1] - y[3]) * ((z * y[1] - y[3])>0) * y[1] * y[4] - chi * (p - y[0]) * ((p - y[0])>0) * y[3], 
            b * (w_bar - y[4]) * ((w_bar - y[4])>0) + alpha * y[5] - (phi * (z * y[0] - y[2]) * ((z * y[0] - y[2])>0) * y[0]\
                + phi * (z * y[1] - y[3]) * ((z * y[1] - y[3])>0) * y[1]) * y[4],
            mu_fct(t) * y[2] - k * y[5] * y[6] - (alpha + delta) * y[5] + gamma * (R_ges - y[6]),
            -k * y[6] * y[5] + gamma * (R_ges - y[6])]
        #return [chi * y[1] - mu_fct(t) * y[0],
        #    - chi * y[1],
        #    chi * y[3] - mu_fct(t) * y[2] + phi_p * y[0] * y[4],
        #    phi_r * y[1] * y[4] - chi * y[3], 
        #    b * (w_bar - y[4]) * ((w_bar - y[4])>0) + alpha * y[5] - (phi_p * y[0] + phi_r * y[1]) * y[4],
        #    mu_fct(t) * y[2] - k * y[5] * y[6] - (alpha + delta) * y[5] + gamma * (R_ges - y[6]),
        #    -k * y[6] * y[5] + gamma * (R_ges - y[6])]

    #def f_jacobian(self, y, t, params): #FIXME: not correct yet
    #    '''Compute the Jacobian (w.r.t. y) of the time derivative function at time t.
    #    Given the parameters params and current system state y.

    #    NOTE: The Jacobian is not correct if time delay is used!'''
    #    assert len(params)>=len(self.keys)

    #    # Model parameters
    #    b, w_bar, alpha, delta, k, gamma, R_ges, chi, phi, z, p = params[:11]

    #    # Function parameters
    #    mu_factor, ca_jump, ca_decay, ca_delay = params[11:15]

    #    ca_baseline = 0.
    #    mu_fct = self.mu(mu_factor, ca_decay, ca_jump, ca_baseline, ca_delay)

    #    return ([   [-mu_fct(t), chi, 0, 0, 0, 0, 0],
    #                [0, -chi, 0, 0, 0, 0, 0],
    #                [phi_p * y[4], 0, -mu_fct(t), chi, phi_p * y[0], 0, 0],
    #                [0, phi_r * y[4], 0, -chi, phi_r * y[1], 0, 0],
    #                [-phi_p * y[4], -phi_r * y[4], 0, 0, -b - phi_p * y[0] - phi_r * y[1], alpha, 0],
    #                [0, 0, mu_fct(t), 0, 0, -k*y[6] - alpha - delta, -k*y[5] - gamma],
    #                [0, 0, 0, 0, 0, -k*y[6], -k*y[5] - gamma]] if y[4] < w_bar else
    #            [   [-mu_fct(t), chi, 0, 0, 0, 0, 0],
    #                [0, -chi, 0, 0, 0, 0, 0],
    #                [phi_p * y[4], 0, -mu_fct(t), chi, phi_p * y[0], 0, 0],
    #                [0, phi_r * y[4], 0, -chi, phi_r * y[1], 0, 0],
    #                [-phi_p * y[4], -phi_r * y[4], 0, 0, - phi_p * y[0] - phi_r * y[1], alpha, 0],
    #                [0, 0, mu_fct(t), 0, 0, -k*y[6] - alpha - delta, -k*y[5] - gamma],
    #                [0, 0, 0, 0, 0, -k*y[6], -k*y[5] - gamma]])


    def solve_system(self, *params):
        '''Calculate the behaviour of all variables for a given input.
        (Note that this function uses global variables for boundary conditions.)

        @param data input data as an array where data[i] contains the input at time t_i
        @param params parameter values for the model
        @return array with the values of the system variables'''        
        # Initialization
        output = zeros((len(self.data),self.variable_number))

        # Applying the boundary conditions
	R_0 = self.paramListToDic(params)['R_ges']
        if len(params) >= len(self.keys) + len(self.boundary_keys):
            x = array(list(params[-len(self.boundary_keys):]) + [R_0])
        else:
            x = array([self.n_p_0, self.n_r_0, self.N_p_0, self.N_r_0, self.N_in_0, self.N_out_0, R_0])

        #return odeint((lambda y,t: self.f(y,t,params)), x, range(len(self.data)), Dfun=(lambda y,t: self.f_jacobian(y,t,params)), **self.odekwargs)
        return odeint((lambda y,t: self.f(y,t,params)), x, range(len(self.data)), **self.odekwargs)


    #=======================================================================================================================
    # EVALUATION
    # Computing how well the model output fits the given data.
    #=======================================================================================================================

    def error(self, params = None):
        '''Calculate the objective function value given a certain configuration of model parameters.

        @param params parameter values for the model
        @param oneDist a function taking two vectors as arguments and returning the distance between them
        @return error for the given parameters'''
        if params is None:
            params = self.valueDicToList(self.p)
        return self.distance(self.output(*params), self.target)


    #=======================================================================================================================
    # VISUALIZATION
    # Displaying output and behaviour of individual variables.
    #=======================================================================================================================

    def plotPerformance(self, params = None):
        '''Plot the input, output and desired output for the given choice of parameters. Note that the input is scaled.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        plot(self.data * amax(self.target)/amax(self.data), label="input", alpha=.6)
        plot(self.output(*params), label="output")
        plot(self.target, label="target")
        legend()

    def plotVariableBehaviour(self, params = None):
        '''Plot the behaviour of the different system variables, given some input and a list of parameters.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        variables = self.solve_system(*params)
        paramDic = self.paramListToDic(params)
        ca_baseline = 0.
        plot([self.calciumConcentration(t, paramDic['ca_decay'], paramDic['ca_jump'], ca_baseline,
            paramDic['ca_delay']) for t in range(len(self.data))], label="$Ca^{2+}$")
        #n_p, n_r, N_p, N_r, N_in, N_out, R, m
        plot(variables[:,0], label="$n_p$")
        plot(variables[:,1], label="$n_r$")
        plot(variables[:,2], label="$N_p$")
        plot(variables[:,3], label="$N_r$")
        plot(variables[:,4], label="$N_{in}$")
        plot(variables[:,5], label="$N_{out}$")
        plot(variables[:,6], label="$R$")
        legend()

    def plotBehaviour(self, params = None):
        '''Plot performance of the model and variable behaviour.

        @param params parameter values for the model'''
        if params is None:
            params = self.valueDicToList(self.p)
        figsize(15, 5)
        subplot(1, 2, 1)
        self.plotPerformance(params)

        subplot(1, 2, 2)
        self.plotVariableBehaviour(params)

        # set figsize back to normal for later plots
        figsize(9,5)


    #=======================================================================================================================
    # OPTIMIZATION
    # Finding optimal parameters for the model.
    #=======================================================================================================================

    def optimize(self, display = True, **kwargs):
        '''Find the optimal parameters.

        @param display If set to True, the optimal parameters are printed out and the performance is plotted.
        @return the (hopefully) optimal parameters'''
        if len(self.p) >= len(self.keys) + len(self.boundary_keys):
            b = self.valueDicToList(self.bounds)
        else:
            b = self.paramDicToList(self.bounds)
        result = fmin_l_bfgs_b((lambda params: self.error(params)), self.valueDicToList(self.p), bounds=b,
            approx_grad=True, **kwargs)
        opt_p = result[0]

        if display:
            if len(self.p) >= len(self.keys) + len(self.boundary_keys):
                print 'optimal parameters:', self.valueListToDic(opt_p)
            else:
                print 'optimal parameters:', self.paramListToDic(opt_p)

            figsize(15, 5)
            subplot(1, 2, 1)
            self.plotPerformance(opt_p)

            subplot(1, 2, 2)
            self.plotVariableBehaviour(opt_p)

            # set figsize back to normal for later plots
            figsize(9,5)

        return result
